#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>

#include "forms.h"
#include "Xsoxdefs.h"

/* global static variables */

char *AudioFormats[] = {
   "auto", "aiff", "au" , "hcom", "ub" , "sb" , "uw" , "sw",
   "ul"  , "sf"  , "voc", "cdr" , "dat", "smp", "wav"
};

static char  cmdbuffer[MAXSTRING];
static pid_t pid;

/* several helper routines */

void AddMessages(FL_OBJECT *ob, long pipe)
{  char message[MAXLINE],*ptr;
   static int count=1;
   
   fl_freeze_form(fd_xsox->Xsox); /* avoid heavy system load */

   ptr=message;
   while(read(pipe,ptr,1) > 0)
   {  count++;
      
      if(*ptr == '\n' || count >= MAXLINE - 1 )
      {
	 *++ptr=0;
	 count=1;
	 fl_addto_browser(fd_xsox->Browser,message);
	 *message=0;
	 ptr=message;
      }
      else ptr++;
   }
   
   *ptr=0;
   fl_addto_browser(fd_xsox->Browser,message);

   /* Clear toplines regularly to avoid heavy memory use */
   if(fl_get_browser_maxline(fd_xsox->Browser) > MAXBROWSER)
   {
      for(count=1;count< (MAXBROWSER / 2);count++)
      	fl_delete_browser_line(fd_xsox->Browser,count);
   }
   
   fl_unfreeze_form(fd_xsox->Xsox);
}

/* This function converts a string into an array like 'argv' , as is used
 * to start a subprocess with exec.
 */
char **BuildArgv(char *string)
{
   int x,count=0;
   char **argv,*ptr;
   
   /* first count the number of arguments */
   
   ptr = string;
   while(*ptr != '\0')
   {
      while (*ptr != ' ' && *ptr != '\0') ptr++;
      while(*ptr == ' ') ptr++; /* skip spaces */
      count ++;
   }
   count++; /* extra space for NULL pointer */

  /* And turn them into an array of separate strings */
   argv = (char **)malloc(count * sizeof(char *));

   argv[0] = ptr = string;
   for(x=1;x<count - 1;x++)
   {
      while (*ptr != ' ') ptr++;
      *ptr++ ='\0';
      while(*ptr == ' ') ptr++; /* skip spaces */
      argv[x]=ptr;
   }
   argv[x]=NULL;
   return argv;
}

/* This forks a process that will i.e. play the audio data or start a separate
 * 'sox' process. This is a little more complicated than just 'system()' as
 * we should be able to kill the playing process through the STOP button
 * and also be able to "chain" them into a pipe sequence by connecting
 * stdin/stdout channels.
 */ 

/* mode OUTPUTREDIR : connect supplied output to stdout of new process
 * mode INPUTREDIR  : connect the stdin of new process to a pipe and return
 *                    the other end of the pipe.
 */

int ForkCommand(char *string,int output,int mode)
{
   int pipepair[2];
   char **argv;
   
   fprintf(stderr,"Executing: %s\n",string);
   argv = BuildArgv(string);  

   if(mode & INPUTREDIR)
      pipe(pipepair);
   
   if ((pid=fork()) == 0) /* store pid to be able to stop playing */
   {
      if( mode & INPUTREDIR)
      {
	 close(STDIN_FILENO);

	 dup(pipepair[0]);
	 close(pipepair[1]);
      }
      if( mode & OUTPUTREDIR )
      {
	 close(STDOUT_FILENO);
	 dup(output);
      }
      
      execvp(argv[0],argv);
      exit(0);
   }
   signal(SIGCHLD,SIG_IGN);
   free(argv);

   /* avoid hanging processes by closing filedescriptors in the parent */
   if( mode & OUTPUTREDIR )
     close(output);
   if(mode & INPUTREDIR)
   {  
      close(pipepair[0]);
      return pipepair[1];
   }
   return 0;
}

/* callbacks for form xsox */

void Convert(FL_OBJECT *ob, long data)
{
   char outputfile[MAXSTRING]="";
   int output,newoutput;
   
   sprintf(cmdbuffer,"sox ");
   AddAmplification(cmdbuffer);

   if(AddInputFile(cmdbuffer) < 0)
     return;

   if(AddOutputFile(outputfile) < 0)
     return;

   AddOutputSpecs(cmdbuffer); /* all but filename */
   strcat(cmdbuffer," -");

   output = open(outputfile,O_WRONLY|O_CREAT,0644);
   newoutput = AddEffects(output);
   ForkCommand(cmdbuffer,newoutput,OUTPUTREDIR);
   wait(NULL);
   close(newoutput);
   close(output);
}

void Play(FL_OBJECT *ob, long data)
{
   char playcmd[MAXSTRING];
   int output,newoutput;
   
   fl_call_object_callback(fd_xsox->Stop);

   sprintf(cmdbuffer,"sox ");
   AddAmplification(cmdbuffer);
   
   if(AddInputFile(cmdbuffer) < 0)
     return;
   
   /* set a default output type if no user defined type specified */
   if(fl_get_choice(fd_xsox->OutputFormat) == 1)
     fl_set_choice(fd_xsox->OutputFormat,PLAYTYPE);
   
   AddOutputSpecs(cmdbuffer); /* all but filename */
   
   strcat(cmdbuffer," -");
   
   strncpy(playcmd,(char *)fl_get_input(fd_xsox->PlayCommand),MAXSTRING - 1);

   output = ForkCommand(playcmd,0,INPUTREDIR);
   newoutput = AddEffects(output);
   ForkCommand(cmdbuffer,newoutput,OUTPUTREDIR);
   close(newoutput);
   close(output);
}

void Stop(FL_OBJECT *ob, long data)
{
   if(pid && ! kill(pid,SIGUNUSED))
   {
      kill(pid,SIGKILL); /* kind of rough, but  ... */
      fl_addto_browser(fd_xsox->Browser,
	"Player terminated, but music may continue until buffer is empty !");
   }
   pid = 0;
}

void InputFS(FL_OBJECT *ob, long data)
{  char *file;
   
   file = (char *)fl_show_fselector("Select Inputfile","","*","");
   if (file != NULL) fl_set_input(fd_xsox->SetInput,file);
}

void OutputFS(FL_OBJECT *ob, long data)
{  char *file;
   
   file = (char *)fl_show_fselector("Select Outputfile","","*","");
   if (file != NULL)
   {
      fl_set_input(fd_xsox->SetOutput,file);    
      fl_set_choice(fd_xsox->OutputFormat,1); /* set to filetype */
   }
}

void Exit(FL_OBJECT *ob, long data)
{
   if(fl_show_question("Do you want to quit ?","",""))
     exit(0);
}

void Info(FL_OBJECT *ob, long data)
{
   fl_show_message("Xsox: a powerful Sound Converter build using the Xforms library",
		   "Sox: by Jef Poskanzer (C) 1989; Xforms: by T. Zhao and M. Overmars",
		   "Xsox: by Kees Lemmens, The Netherlands (C) 1996" VERSION);
}
